"""Implements class converts data from and to np.ndarray"""

from __future__ import absolute_import

import logging

import numpy as np

logger = logging.getLogger(__name__)


# pylint: disable=unused-argument
class TorchTensorTranslator:
    """Translate torch.Tensor from and to numpy.ndarray"""

    def __init__(self) -> None:
        import torch

        self.convert_from_numpy = torch.from_numpy  # pylint: disable=E1101
        self.CONTENT_TYPE = "tensor/pt"
        self.ACCEPT = "tensor/pt"

    def serialize(self, data, content_type: str = "tensor/pt"):
        """Translate torch.Tensor to numpy ndarray"""
        try:
            return data.detach().numpy()
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to translate data %s to np.ndarray: %s" % (type(data), e))

    def deserialize(self, data, content_type: str = "application/x-npy"):
        """Translate numpy ndarray to torch.Tensor"""
        try:
            return self.convert_from_numpy(data)
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to translate data %s to torch.Tensor: %s " % (type(data), e))

    def _deserializer(self):
        """Dummy function to align with DeserializerWrapper in SchemaBuilder"""
        raise ValueError("This method is not meant to be invoked.")


class TensorflowTensorTranslator:
    """Converts tf.Tensor from and to numpy.ndarray"""

    def __init__(self) -> None:
        import tensorflow as tf

        self.convert_to_tensor = tf.convert_to_tensor
        self.CONTENT_TYPE = "tensor/tf"
        self.ACCEPT = "tensor/tf"

    def serialize(self, data, content_type: str = "tensor/tf"):
        """Translate tf.Tensor to numpy ndarray"""
        try:
            return data.numpy()
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to convert data %s to np.ndarray" % type(data)) from e

    def deserialize(self, data, content_type: str = "application/x-npy"):
        """Translate numpy ndarray to torch.Tensor"""
        try:
            return self.convert_to_tensor(data)
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to convert data %s to tf.Tensor" % type(data)) from e

    def _deserializer(self):
        """Dummy function to align with DeserializerWrapper in SchemaBuilder"""
        raise ValueError("This method is not meant to be invoked.")


class NumpyTranslator:
    """A dummy class to make sure the translator interface is aligned"""

    def __init__(self) -> None:
        self.CONTENT_TYPE = "application/x-npy"
        self.ACCEPT = "application/x-npy"

    def serialize(self, data, content_type: str = "application/x-npy"):
        """Placeholder docstring"""
        return data

    def deserialize(self, data, content_type: str = "application/x-npy"):
        """Placeholder docstring"""
        return data

    def _deserializer(self):
        """Dummy function to align with DeserializerWrapper in SchemaBuilder"""
        raise ValueError("This method is not meant to be invoked.")


class ListTranslator:
    """Translate python list from and to numpy.ndarray"""

    def __init__(self) -> None:
        self.CONTENT_TYPE = "application/list"
        self.ACCEPT = "application/list"

    def serialize(self, data, content_type: str = "application/list"):
        """Placeholder docstring"""
        try:
            return np.array(data)
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to convert data %s to np.ndarray" % type(data)) from e

    def deserialize(self, data, content_type: str = "application/x-npy"):
        """Placeholder docstring"""
        try:
            return data.tolist()
        except Exception as e:
            logger.error(e)
            raise ValueError("Unable to convert data %s to python list" % type(data)) from e

    def _deserializer(self):
        """Dummy function to align with DeserializerWrapper in SchemaBuilder"""
        raise ValueError("This method is not meant to be invoked.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Accessors to retrieve task fallback input/output schema"""
from __future__ import absolute_import

import json
import os
from typing import Any, Tuple


def retrieve_local_schemas(task: str) -> Tuple[Any, Any]:
    """Retrieves task sample inputs and outputs locally.

    Args:
        task (str): Required, the task name

    Returns:
        Tuple[Any, Any]: A tuple that contains the sample input,
        at index 0, and output schema, at index 1.

    Raises:
        ValueError: If no tasks config found or the task does not exist in the local config.
    """
    config_dir = os.path.dirname(os.path.dirname(__file__))
    task_io_config_path = os.path.join(config_dir, "schema", "task.json")
    try:
        with open(task_io_config_path) as f:
            task_io_config = json.load(f)
            task_io_schemas = task_io_config.get(task, None)

            if task_io_schemas is None:
                raise ValueError(f"Could not find {task} I/O schema.")

            sample_schema = (
                task_io_schemas["sample_inputs"]["properties"],
                task_io_schemas["sample_outputs"]["properties"],
            )
        return sample_schema
    except FileNotFoundError:
        raise ValueError("Could not find tasks config file.")



"""Upload model artifacts to S3"""

from __future__ import absolute_import
import logging
import os
import tempfile
import botocore
import boto3
import tqdm
from sagemaker.core.helper.session_helper import Session
from sagemaker.core.s3.utils import s3_path_join
from sagemaker.core.s3 import S3Uploader
from sagemaker.core.common_utils import create_tar_file

logger = logging.getLogger(__name__)

# Minimum size required for multi-part uploads
BUF_SIZE = 5 * 1024 * 1024


def _get_dir_size(path):
    """Calculate the size of a directory"""
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += _get_dir_size(entry.path)
    return total


class Uploader(object):
    """Uploader class that handles uploading data to S3 and display progress bar"""

    def __init__(self) -> None:
        self.total_left = None
        self.pbar = None

    def observe(self, bytes_amount):
        """Placeholder docstring"""
        self.total_left -= bytes_amount
        self.pbar.update(bytes_amount)

    def upload(
        self,
        model_dir,
        total_size: int,
        credentials: botocore.credentials.Credentials,
        region_name: str,
        bucket: str,
        key: str,
    ):
        """Compress and upload the model tar object to S3"""
        self.total_left = total_size
        with tqdm.tqdm(
            total=total_size, desc="Uploading model artifacts", unit="bytes", ncols=100
        ) as self.pbar:
            files = [os.path.join(model_dir, name) for name in os.listdir(model_dir)]
            temp = tempfile.mkdtemp()
            tar_file = create_tar_file(files, os.path.join(temp, "model.tar.gz"))

            s3 = boto3.session.Session(
                region_name=region_name,
                aws_access_key_id=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                aws_session_token=credentials.token,
            ).client("s3")
            s3.upload_file(os.path.join(temp, "model.tar.gz"), bucket, key, Callback=self.observe)
            os.remove(tar_file)
            self.pbar.update(self.total_left)
            self.pbar.close()
            self.pbar = None

    def upload_uncompressed(
        self,
        model_dir: str,
        sagemaker_session: Session,
        bucket: str,
        key_prefix: str,
        total_size: int,
    ):
        """Upload uncompressed model artifacts to S3"""
        self.total_left = total_size
        with tqdm.tqdm(
            total=total_size, desc="Uploading model artifacts", unit="bytes", ncols=100
        ) as self.pbar:
            S3Uploader.upload(
                local_path=model_dir,
                desired_s3_uri=s3_path_join("s3://", bucket, key_prefix),
                sagemaker_session=sagemaker_session,
                callback=self.observe,
            )
            self.pbar.update(self.total_left)


def upload(sagemaker_session: Session, model_dir: str, bucket: str, key_prefix: str) -> str:
    """Wrapper function of method upload"""
    key = key_prefix + "/serve.tar.gz"
    uploader = Uploader()
    uploader.upload(
        model_dir,
        _get_dir_size(model_dir),
        sagemaker_session.boto_session.get_credentials(),
        sagemaker_session.boto_session.region_name,
        bucket,
        key,
    )
    return s3_path_join("s3://", bucket, key)


def upload_uncompressed(
    sagemaker_session: Session, model_dir: str, bucket: str, key_prefix: str
) -> str:
    """Wrapper function of method upload_uncompressed"""
    uploader = Uploader()
    uploader.upload_uncompressed(
        model_dir, sagemaker_session, bucket, key_prefix, _get_dir_size(model_dir)
    )
    return s3_path_join("s3://", bucket, key_prefix, with_end_slash=True)



"""Placeholder docstring"""

from __future__ import absolute_import
import io
import logging
from pathlib import Path
from typing import Callable
import numpy as np
from pandas import DataFrame

from sagemaker.core.deserializers import (
    BaseDeserializer,
    BytesDeserializer,
    NumpyDeserializer,
    JSONDeserializer,
    PandasDeserializer,
    TorchTensorDeserializer,
    StringDeserializer,
)
from sagemaker.core.serializers import (
    DataSerializer,
    NumpySerializer,
    JSONSerializer,
    CSVSerializer,
    TorchTensorSerializer,
    StringSerializer,
)

from sagemaker.serve.marshalling.custom_payload_translator import CustomPayloadTranslator

from sagemaker.serve.builder.triton_schema_builder import TritonSchemaBuilder

logger = logging.getLogger(__name__)


class JSONSerializerWrapper(JSONSerializer):
    """Wraps the JSONSerializer because it does not convert jsonable to bytes"""

    def serialize(self, data) -> bytes:
        """Placeholder docstring"""

        return super().serialize(data).encode("utf-8")


class CSVSerializerWrapper(CSVSerializer):
    """Wraps the CSVSerializer because it does not convert dataframe to bytes"""

    def serialize(self, data) -> bytes:
        """Placeholder docstring"""
        return super().serialize(data).encode("utf-8")


translation_mapping = {
    NumpySerializer: NumpyDeserializer,
    NumpyDeserializer: NumpySerializer,
    JSONSerializerWrapper: JSONDeserializer,
    JSONDeserializer: JSONSerializerWrapper,
    TorchTensorSerializer: TorchTensorDeserializer,
    TorchTensorDeserializer: TorchTensorSerializer,
    DataSerializer: BytesDeserializer,
    BytesDeserializer: DataSerializer,
    CSVSerializerWrapper: PandasDeserializer,
    PandasDeserializer: CSVSerializerWrapper,
    StringSerializer: StringDeserializer,
    StringDeserializer: StringSerializer,
}


class DeserializerWrapper(BaseDeserializer):
    """Wraps the deserializer to comply with the function signature."""

    def __init__(self, deserializer, accept):
        self._deserializer = deserializer
        self._accept = accept

    def deserialize(self, stream, content_type: str = None):
        """Deserialize stream into object"""
        return self._deserializer.deserialize(
            stream,
            # We need to overwrite the accept type because model
            # servers like XGBOOST always returns "text/html"
            self._accept[0],
        )

    @property
    def ACCEPT(self):
        """Placeholder docstring"""
        return self._accept[0]


class SchemaBuilder(TritonSchemaBuilder):
    """Automatically detects the serializer and deserializer for your model.

    This is done by inspecting the `sample_input` and `sample_output` object.
    Alternatively, provide your custom serializer and deserializer
    for your request or response by creating a class that inherits
    ``CustomPayloadTranslator`` and provide it to ``SchemaBuilder``.

    Args:
       sample_input (object): Sample input to the model which can be used
           for testing. The schema builder internally generates the content
           type and corresponding serializing functions.
       sample_output (object): Sample output to the model which can be
           used for testing. The schema builder internally generates
           the accept type and corresponding serializing functions.
       input_translator (Optional[CustomPayloadTranslator]): If you
           want to define your own serialization method for the payload,
           you can implement your functions for translation.
       output_translator (Optional[CustomPayloadTranslator]): If
           you want to define your own serialization method for the output,
           you can implement your functions for translation.
    """

    def __init__(
        self,
        sample_input,
        sample_output,
        input_translator: CustomPayloadTranslator = None,
        output_translator: CustomPayloadTranslator = None,
    ):
        super().__init__()

        self.sample_input = sample_input
        self.sample_output = sample_output
        if input_translator:
            _validate_translations(
                payload=sample_input,
                serialize_callable=input_translator.serialize,
                deserialize_callable=input_translator.deserialize,
            )
            self.custom_input_translator = input_translator
        else:
            self.input_serializer = self._get_serializer(sample_input)
            self._input_deserializer = self._get_inverse(self.input_serializer)
            self.input_deserializer = DeserializerWrapper(
                self._input_deserializer, self._input_deserializer.ACCEPT
            )

        if output_translator:
            _validate_translations(
                payload=sample_output,
                serialize_callable=output_translator.serialize,
                deserialize_callable=output_translator.deserialize,
            )
            self.custom_output_translator = output_translator
        else:
            self._output_deserializer = self._get_deserializer(sample_output)
            self.output_serializer = self._get_inverse(self._output_deserializer)

            self.output_deserializer = DeserializerWrapper(
                self._output_deserializer, self._output_deserializer.ACCEPT
            )

    def _get_serializer(self, obj):
        # pylint: disable=too-many-return-statements
        """Placeholder docstring"""
        if isinstance(obj, np.ndarray):
            return NumpySerializer()
        if isinstance(obj, DataFrame):
            return CSVSerializerWrapper()
        if isinstance(obj, bytes) or _is_path_to_file(obj):
            return DataSerializer()
        if _is_torch_tensor(obj):
            return TorchTensorSerializer()
        if isinstance(obj, str):
            return StringSerializer()
        if _is_jsonable(obj):
            return JSONSerializerWrapper()
        if isinstance(obj, dict) and "content_type" in obj:
            try:
                return DataSerializer(content_type=obj["content_type"])
            except ValueError as e:
                logger.error(e)

        raise ValueError(
            (
                "SchemaBuilder cannot determine the serializer of type %s "
                "Please provide your own marshalling code"
                "to SchemaBuilder via CustomPayloadTranslator"
            )
            % type(obj)
        )

    def _get_deserializer(self, obj):
        # pylint: disable=too-many-return-statements
        """Placeholder docstring"""
        if isinstance(obj, np.ndarray):
            return NumpyDeserializer()
        if isinstance(obj, DataFrame):
            return PandasDeserializer()
        if isinstance(obj, bytes):
            return BytesDeserializer()
        if _is_torch_tensor(obj):
            return TorchTensorDeserializer()
        if isinstance(obj, str):
            return StringDeserializer()
        if _is_jsonable(obj):
            return JSONDeserializer()

        raise ValueError(
            (
                "SchemaBuilder cannot determine deserializer of type %s "
                "Please provide your own marshalling code"
                "to SchemaBuilder via CustomPayloadTranslator"
            )
            % type(obj)
        )

    def _get_inverse(self, obj):
        """Placeholder docstring"""
        try:
            return translation_mapping.get(obj.__class__)()
        except KeyError:
            raise Exception("Unable to serialize")

    def __repr__(self):
        """Placeholder docstring"""
        if hasattr(self, "input_serializer") and hasattr(self, "output_serializer"):
            return (
                f"SchemaBuilder(\n"
                f"input_serializer={self.input_serializer}\n"
                f"output_serializer={self.output_serializer}\n"
                f"input_deserializer={self.input_deserializer._deserializer}\n"
                f"output_deserializer={self.output_deserializer._deserializer})"
            )
        return (
            f"SchemaBuilder(\n"
            f"custom_input_translator={self.custom_input_translator}\n"
            f"custom_output_translator={self.custom_output_translator}\n"
        )

    def generate_marshalling_map(self) -> dict:
        """Generate marshalling map for the schema builder"""
        return {
            "input_serializer": (
                self.input_serializer.__class__.__name__
                if hasattr(self, "input_serializer")
                else None
            ),
            "output_serializer": (
                self.output_serializer.__class__.__name__
                if hasattr(self, "output_serializer")
                else None
            ),
            "input_deserializer": (
                self._input_deserializer.__class__.__name__
                if hasattr(self, "_input_deserializer")
                else None
            ),
            "output_deserializer": (
                self._output_deserializer.__class__.__name__
                if hasattr(self, "_output_deserializer")
                else None
            ),
            "custom_input_translator": hasattr(self, "custom_input_translator"),
            "custom_output_translator": hasattr(self, "custom_output_translator"),
        }

    def get_input_sample(self) -> object:
        """Get input sample for the schema builder"""
        return self.sample_input


def _is_torch_tensor(data: object) -> bool:
    """Placeholder docstring"""
    try:
        from torch import Tensor

        return isinstance(data, Tensor)
    except ModuleNotFoundError:
        return False


def _is_jsonable(data: object) -> bool:
    # pylint: disable=broad-except
    """Placeholder docstring"""
    try:
        JSONSerializerWrapper().serialize(data)
        return True
    except Exception:
        return False


def _is_path_to_file(data: object) -> bool:
    """Placeholder docstring"""
    return isinstance(data, str) and Path(data).resolve().is_file()


def _validate_translations(
    payload: object, serialize_callable: Callable, deserialize_callable: Callable
) -> None:
    """Placeholder docstring"""
    try:
        b = serialize_callable(payload=payload, content_type="application/custom")
        stream = io.BytesIO(b)
        deserialize_callable(stream=stream, content_type="application/custom")
    except Exception as e:
        raise ValueError("Error when validating payload serialization and deserialization.", e)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Helper classes that handles intelligent default values for serve function"""
from __future__ import absolute_import
from typing import Dict

from sagemaker.core.helper.session_helper import Session
from sagemaker.core.common_utils import resolve_value_from_config

from sagemaker.core.config.config_schema import (
    SERVE_ENVIRONMENT_VARIABLES,
    SERVE_INSTANCE_TYPE,
    SERVE_ROLE_ARN,
    SERVE_S3_MODEL_DATA_URI,
    TELEMETRY_OPT_OUT_PATH,
)


class _ServeSettings(object):
    """Helper class that processes the job settings.

    It validates the job settings and provides default values if necessary.
    """

    def __init__(
        self,
        role_arn: str = None,
        s3_model_data_url: str = None,
        instance_type: str = None,
        env_vars: Dict[str, str] = None,
        sagemaker_session: Session = None,
    ):
        self.sagemaker_session = sagemaker_session or Session()

        self.role_arn = resolve_value_from_config(
            direct_input=role_arn,
            config_path=SERVE_ROLE_ARN,
            sagemaker_session=self.sagemaker_session,
        )
        self.s3_model_data_url = resolve_value_from_config(
            direct_input=s3_model_data_url,
            config_path=SERVE_S3_MODEL_DATA_URI,
            sagemaker_session=self.sagemaker_session,
        )
        self.instance_type = resolve_value_from_config(
            direct_input=instance_type,
            config_path=SERVE_INSTANCE_TYPE,
            sagemaker_session=self.sagemaker_session,
        )
        self.env_vars = resolve_value_from_config(
            direct_input=env_vars,
            config_path=SERVE_ENVIRONMENT_VARIABLES,
            default_value={},
            sagemaker_session=self.sagemaker_session,
        )
        self.telemetry_opt_out = resolve_value_from_config(
            direct_input=None,
            config_path=TELEMETRY_OPT_OUT_PATH,
            default_value=False,
            sagemaker_session=self.sagemaker_session,
        )



"""Placeholder docstring"""

from __future__ import absolute_import
from enum import Enum

PKL_FILE_NAME = "serve.pkl"


class Mode(Enum):
    """Placeholder docstring"""

    def __str__(self):
        """Placeholder docstring"""
        return str(self.name)

    IN_PROCESS = 1
    LOCAL_CONTAINER = 2
    SAGEMAKER_ENDPOINT = 3



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Prepare DjlModel for Deployment"""

from __future__ import absolute_import
import json
import tarfile
import logging
from typing import List
from pathlib import Path

from sagemaker.core.common_utils import _tmpdir, custom_extractall_tarfile
from sagemaker.core.s3 import S3Downloader
from sagemaker.serve.utils.local_hardware import _check_disk_space, _check_docker_disk_usage

_SETTING_PROPERTY_STMT = "Setting property: %s to %s"

logger = logging.getLogger(__name__)


def _extract_js_resource(js_model_dir: str, code_dir: Path, js_id: str):
    """Uncompress the jumpstart resource"""
    tmp_sourcedir = Path(js_model_dir).joinpath(f"infer-prepack-{js_id}.tar.gz")
    with tarfile.open(str(tmp_sourcedir)) as resources:
        custom_extractall_tarfile(resources, code_dir)


def _copy_jumpstart_artifacts(model_data: str, js_id: str, code_dir: Path) -> tuple:
    """Copy the associated JumpStart Resource into the code directory"""
    logger.info("Downloading JumpStart artifacts from S3...")

    s3_downloader = S3Downloader()
    if isinstance(model_data, str):
        if model_data.endswith(".tar.gz"):
            logger.info("Uncompressing JumpStart artifacts for faster loading...")
            with _tmpdir(directory=str(code_dir)) as js_model_dir:
                s3_downloader.download(model_data, js_model_dir)
                _extract_js_resource(js_model_dir, code_dir, js_id)
        else:
            logger.info("Copying uncompressed JumpStart artifacts...")
            s3_downloader.download(model_data, code_dir)
    elif (
        isinstance(model_data, dict)
        and model_data.get("S3DataSource")
        and model_data.get("S3DataSource").get("S3Uri")
    ):
        logger.info("Copying uncompressed JumpStart artifacts...")
        s3_downloader.download(model_data.get("S3DataSource").get("S3Uri"), code_dir)
    else:
        raise ValueError("JumpStart model data compression format is unsupported: %s", model_data)

    config_json_file = code_dir.joinpath("config.json")
    hf_model_config = None
    if config_json_file.is_file():
        with open(str(config_json_file)) as config_json:
            hf_model_config = json.load(config_json)

    return (hf_model_config, True)


def _create_dir_structure(model_path: str) -> tuple:
    """Placeholder Docstring"""
    model_path = Path(model_path)
    if not model_path.exists():
        model_path.mkdir(parents=True)
    elif not model_path.is_dir():
        raise ValueError("model_dir is not a valid directory")

    code_dir = model_path.joinpath("code")
    code_dir.mkdir(exist_ok=True, parents=True)

    _check_disk_space(model_path)
    _check_docker_disk_usage()

    return (model_path, code_dir)


def prepare_djl_js_resources(
    model_path: str,
    js_id: str,
    shared_libs: List[str] = None,
    dependencies: str = None,
    model_data: str = None,
) -> tuple:
    """Prepare serving when a JumpStart model id is given

    Args:
        model_path (str) : Argument
        js_id (str): Argument
        to_uncompressed (bool): Argument
        shared_libs (List[]) : Argument
        dependencies (str) : Argument
        model_data (str) : Argument

    Returns:
        ( str ) :

    """
    model_path, code_dir = _create_dir_structure(model_path)

    return _copy_jumpstart_artifacts(model_data, js_id, code_dir)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""SageMaker model builder dependency managing module.

This must be kept independent of SageMaker PySDK
"""

from __future__ import absolute_import

from pathlib import Path
import logging
import subprocess
import sys
import re

_SUPPORTED_SUFFIXES = [".txt"]
# TODO : Move PKL_FILE_NAME to common location
PKL_FILE_NAME = "serve.pkl"

logger = logging.getLogger(__name__)


def capture_dependencies(dependencies: dict, work_dir: Path, capture_all: bool = False):
    """Placeholder docstring"""
    path = work_dir.joinpath("requirements.txt")
    if "auto" in dependencies and dependencies["auto"]:
        import site

        pkl_path = work_dir.joinpath(PKL_FILE_NAME).resolve()
        dest_path = path.resolve()
        site_packages_dir = site.getsitepackages()[0]
        pickle_command_dir = "/sagemaker/serve/detector"

        command = [
            sys.executable,
            "-c",
        ]

        if capture_all:
            command.append(
                f"from sagemaker.serve.detector.pickle_dependencies import get_all_requirements;"
                f'get_all_requirements("{dest_path}")'
            )
        else:
            command.append(
                f"from sagemaker.serve.detector.pickle_dependencies import get_requirements_for_pkl_file;"
                f'get_requirements_for_pkl_file("{pkl_path}", "{dest_path}")'
            )

        subprocess.run(
            command,
            env={"SETUPTOOLS_USE_DISTUTILS": "stdlib"},
            check=True,
            cwd=site_packages_dir + pickle_command_dir,
        )

        with open(path, "r") as f:
            autodetect_depedencies = f.read().splitlines()
        # autodetect_depedencies.append("sagemaker[huggingface]>=2.199")
        # autodetect_depedencies = []
    else:
        autodetect_depedencies = []
        # autodetect_depedencies = ["sagemaker[huggingface]>=2.199"]

    module_version_dict = _parse_dependency_list(autodetect_depedencies)

    if "requirements" in dependencies:
        module_version_dict = _process_customer_provided_requirements(
            requirements_file=dependencies["requirements"], module_version_dict=module_version_dict
        )
    if "custom" in dependencies:
        module_version_dict = _process_custom_dependencies(
            custom_dependencies=dependencies.get("custom"), module_version_dict=module_version_dict
        )
    with open(path, "w") as f:
        for module, version in module_version_dict.items():
            f.write(f"{module}{version}\n")


def _process_custom_dependencies(custom_dependencies: list, module_version_dict: dict):
    """Placeholder docstring"""
    custom_module_version_dict = _parse_dependency_list(custom_dependencies)
    module_version_dict.update(custom_module_version_dict)
    return module_version_dict


def _process_customer_provided_requirements(requirements_file: str, module_version_dict: dict):
    """Placeholder docstring"""
    requirements_file = Path(requirements_file)
    if not requirements_file.is_file() or not _is_valid_requirement_file(requirements_file):
        raise Exception(f"Path: {requirements_file} to requirements.txt doesn't exist")
    logger.debug("Packaging provided requirements.txt from %s", requirements_file)
    with open(requirements_file, "r") as f:
        custom_dependencies = f.read().splitlines()

    module_version_dict.update(_parse_dependency_list(custom_dependencies))
    return module_version_dict


def _is_valid_requirement_file(path):
    """Placeholder docstring"""
    # In the future, we can also check the if the content of customer provided file has valid format
    for suffix in _SUPPORTED_SUFFIXES:
        if path.name.endswith(suffix):
            return True
    return False


def _parse_dependency_list(depedency_list: list) -> dict:
    """Placeholder docstring"""

    # Divide a string into 2 part, first part is the module name
    # and second part is its version constraint or the url
    # checkout tests/unit/sagemaker/serve/detector/test_dependency_manager.py
    # for examples
    pattern = r"^([\w.-]+)(@[^,\n]+|((?:[<>=!~]=?[\w.*-]+,?)+)?)$"

    module_version_dict = {}

    for dependency in depedency_list:
        if dependency.startswith("#"):
            continue
        match = re.match(pattern, dependency)
        if match:
            package = match.group(1)
            # Group 2 is either a URL or version constraint, if present
            url_or_version = match.group(2) if match.group(2) else ""
            module_version_dict.update({package: url_or_version})
        else:
            module_version_dict.update({dependency: ""})

    return module_version_dict



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Requirements Manager class to pull in client dependencies from a .txt or .yml file"""
from __future__ import absolute_import
import logging
import os
import subprocess

from typing import Optional

logger = logging.getLogger(__name__)


class RequirementsManager:
    """Manages dependency installation by detecting file types"""

    def capture_and_install_dependencies(self, dependencies: Optional[str] = None) -> str:
        """Detects the type of file dependencies will be installed from

        If a req.txt or conda.yml file is provided, it verifies their existence and
        returns the local file path

        Args:
            dependencies (str): Local path where dependencies file exists.

        Returns:
            file path of the existing or generated dependencies file
        """
        _dependencies = dependencies or self._detect_conda_env_and_local_dependencies

        # Dependencies specified as either req.txt or conda_env.yml
        if _dependencies.endswith(".txt"):
            self._install_requirements_txt()
        elif _dependencies.endswith(".yml"):
            self._update_conda_env_in_path()
        else:
            raise ValueError(f'Invalid dependencies provided: "{_dependencies}"')

    def _install_requirements_txt(self):
        """Install requirements.txt file using pip"""
        logger.info("Running command to pip install")
        subprocess.run("pip install -r in_process_requirements.txt", shell=True, check=True)
        logger.info("Command ran successfully")

    def _update_conda_env_in_path(self):
        """Update conda env using conda yml file"""
        logger.info("Updating conda env")
        subprocess.run("conda env update -f conda_in_process.yml", shell=True, check=True)
        logger.info("Conda env updated successfully")

    def _get_active_conda_env_name(self) -> str:
        """Returns the conda environment name from the set environment variable. None otherwise."""
        return os.getenv("CONDA_DEFAULT_ENV")

    def _get_active_conda_env_prefix(self) -> str:
        """Returns the conda prefix from the set environment variable. None otherwise."""
        return os.getenv("CONDA_PREFIX")

    def _detect_conda_env_and_local_dependencies(self) -> str:
        """Generates dependencies list from the user's local runtime.

        Raises RuntimeEnvironmentError if not able to.

        Currently supports: conda environments
        """

        # Try to capture dependencies from the conda environment, if any.
        conda_env_name = self._get_active_conda_env_name()
        logger.info("Found conda_env_name: '%s'", conda_env_name)
        conda_env_prefix = None

        if conda_env_name is None:
            conda_env_prefix = self._get_active_conda_env_prefix()

        if conda_env_name is None and conda_env_prefix is None:
            local_dependencies_path = os.path.join(os.getcwd(), "in_process_requirements.txt")
            logger.info(local_dependencies_path)

            return local_dependencies_path

        if conda_env_name == "base":
            logger.warning(
                "We recommend using an environment other than base to "
                "isolate your project dependencies from conda dependencies"
            )

        local_dependencies_path = os.path.join(os.getcwd(), "conda_in_process.yml")
        logger.info(local_dependencies_path)

        return local_dependencies_path



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Local Model and Endpoint classes for V3 ModelBuilder local mode support.

These classes provide sagemaker-core compatible interfaces for local deployment,
wrapping V2 local mode functionality.
"""

from __future__ import absolute_import
import datetime
import json
import logging
from typing import Any, Dict, Optional, Tuple
import io
import json
from sagemaker.serve.utils.types import ModelServer
from sagemaker.core.serializers import JSONSerializer, IdentitySerializer
from sagemaker.core.deserializers import JSONDeserializer, BytesDeserializer
from sagemaker.core.resources import Model

logger = logging.getLogger(__name__)

APPLICATION_X_NPY = "application/x-npy"

# Triton gets serializer/deserializer from schema_builder
DEFAULT_SERIALIZERS_BY_SERVER: Dict[ModelServer, Tuple] = {
    ModelServer.TORCHSERVE: (IdentitySerializer(), BytesDeserializer()),
    ModelServer.TENSORFLOW_SERVING: (JSONSerializer(), JSONDeserializer()),  # TF Serving expects JSON
    ModelServer.DJL_SERVING: (JSONSerializer(), JSONDeserializer()),
    ModelServer.TEI: (JSONSerializer(), JSONDeserializer()),
    ModelServer.TGI: (JSONSerializer(), JSONDeserializer()),
    ModelServer.MMS: (JSONSerializer(), JSONDeserializer()),
    ModelServer.SMD: (JSONSerializer(), JSONDeserializer()),
}


class InvokeEndpointOutput:
    """Response wrapper to match sagemaker-core Endpoint.invoke() output format."""
    
    def __init__(self, body: bytes, content_type: str = "application/json"):
        self.body = body
        self.content_type = content_type

class LocalEndpoint:
    """Local endpoint that mimics sagemaker.core.Endpoint interface.
    
    This class wraps V2 LocalSession endpoint functionality to provide a unified
    interface compatible with sagemaker-core Endpoint resources.
    """
    
    def __init__(
        self,
        endpoint_name: str,
        endpoint_config_name: str,
        local_session=None,
        local_model=None,
        in_process_mode=False, 
        local_container_mode_obj=None,
        in_process_mode_obj=None,
        model_server=None,
        secret_key=None,
        serializer=None,
        deserializer=None,
        container_config="auto",
        **kwargs
    ):
        """Initialize local endpoint.
        
        Args:
            endpoint_name: Name of the endpoint
            endpoint_config_name: Name of the endpoint configuration
            local_session: V2 LocalSession instance
        """
        self.endpoint_name = endpoint_name
        self.endpoint_config_name = endpoint_config_name
        self.creation_time = datetime.datetime.now()
        self._local_model = local_model
        self.in_process_mode = in_process_mode
        self.local_container_mode_obj=local_container_mode_obj
        self.in_process_mode_obj=in_process_mode_obj
        self.model_server=model_server
        self.secret_key=secret_key
        self.serializer=serializer
        self.deserializer=deserializer
        self.container_config=container_config
        
        # Import V3 LocalSession
        if local_session is None:
            from sagemaker.core.local.local_session import LocalSession
            self._local_session = LocalSession()
        else:
            self._local_session = local_session
    
    # @property
    # def endpoint_arn(self) -> str:
    #     """Fake ARN for compatibility with sagemaker-core interface."""
    #     return f"arn:aws:sagemaker:local:000000000000:endpoint/{self.endpoint_name}"
    
    @property
    def endpoint_status(self) -> str:
        """Get endpoint status.
        
        Implementation based on V2 LocalSession.describe_endpoint()
        Reference: /sagemaker/local/local_session.py:describe_endpoint()
        """
        try:
            endpoint_info = self._local_session.sagemaker_client.describe_endpoint(
                EndpointName=self.endpoint_name
            )
            return endpoint_info["EndpointStatus"]
        except Exception:
            return "Failed"
    

    def _universal_deep_ping(self) -> tuple[bool, Any]:
        """Universal ping function that works for all model servers."""
        response = None
        logger.info("Pinging local endpoint...")
        try:
            # Get sample input from schema builder
            if self.in_process_mode:
                sample_input = self.in_process_mode_obj.schema_builder.sample_input
            else:
                sample_input = self.local_container_mode_obj.schema_builder.sample_input
            
            # Use unified invoke interface
            invoke_response = self.invoke(body=sample_input)
            
            if self.in_process_mode:
                # IN_PROCESS: Response is already deserialized
                response = invoke_response.body
                healthy = response is not None
            else:
                # LOCAL_CONTAINER: Response needs decoding
                response_body = invoke_response.body.read().decode('utf-8')
                response = json.loads(response_body)
                healthy = response is not None
            
            return (healthy, response)
            
        except Exception as e:
            if "422 Client Error: Unprocessable Entity for url" in str(e):
                from sagemaker.serve.utils.exceptions import LocalModelInvocationException
                raise LocalModelInvocationException(str(e))
            
            return (False, None)


    def invoke(
        self,
        body: Any,
        content_type: str = "application/json",
        accept: str = "application/json",
        **kwargs
    ) -> InvokeEndpointOutput:
        """Invoke the local endpoint using model server-specific logic."""
        if self.in_process_mode:
            if not self.in_process_mode_obj:
                raise ValueError("In Process container mode not available")
            
            serializer = self.serializer or JSONSerializer()
            deserializer = self.deserializer or JSONDeserializer()
            serialized_data = serializer.serialize(body)

            raw_response = self.in_process_mode_obj._invoke_serving(
                serialized_data, content_type, accept
            )
            return InvokeEndpointOutput(
                body=deserializer.deserialize(io.BytesIO(raw_response)),
                content_type=accept
            )
        
        else:
            if not self.model_server or not self.local_container_mode_obj:
                raise ValueError("Model server or container mode not available")
            
            # Get serializers (use defaults if not provided by model)
            serializer = self.serializer or JSONSerializer()
            deserializer = self.deserializer or JSONDeserializer()

            content_type = content_type if content_type != "application/json" else serializer.CONTENT_TYPE
            deserializer_accept = deserializer.ACCEPT
            if not isinstance(deserializer_accept, str):
                deserializer_accept = deserializer_accept[0]
            accept = accept if accept != "application/json" else deserializer_accept
            
            # Route to appropriate model server invoke method
            if self.model_server == ModelServer.TORCHSERVE:
                # TorchServe: Use serializer-derived content types (V2 pattern)
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_torch_serve(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = deserializer.deserialize(io.BytesIO(raw_response))
                
            elif self.model_server == ModelServer.TRITON:
                # Triton: Direct data, no serialization, fixed content types (V2 pattern)
                from sagemaker.serve.utils.predictors import APPLICATION_X_NPY
                raw_response = self.local_container_mode_obj._invoke_triton_server(
                    body,  # ← Direct data, no serialization
                    APPLICATION_X_NPY,
                    APPLICATION_X_NPY
                )
                response_data = raw_response
                
            elif self.model_server == ModelServer.DJL_SERVING:
                # DJL: Use serializer-derived content types + deserialize with content_type
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_djl_serving(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = deserializer.deserialize(
                    io.BytesIO(raw_response),
                    content_type 
                )
                
            elif self.model_server == ModelServer.TGI:
                # TGI: Use serializer-derived content types + list format
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_tgi_serving(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = [deserializer.deserialize(
                    io.BytesIO(raw_response),
                    content_type
                )]
                
            elif self.model_server == ModelServer.MMS:
                # MMS: Use serializer-derived content types + list format
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_multi_model_server_serving(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = [deserializer.deserialize(
                    io.BytesIO(raw_response),
                    content_type
                )]
                
            elif self.model_server == ModelServer.TENSORFLOW_SERVING:
                # TensorFlow: Use serializer-derived content types
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_tensorflow_serving(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = deserializer.deserialize(io.BytesIO(raw_response))
                
            elif self.model_server == ModelServer.TEI:
                # TEI: Use serializer-derived content types
                serialized_data = serializer.serialize(body) if not isinstance(body, str) else body
                raw_response = self.local_container_mode_obj._invoke_serving(
                    serialized_data,
                    content_type,
                    accept
                )
                response_data = deserializer.deserialize(io.BytesIO(raw_response))
                
            else:
                raise ValueError(f"Unsupported model server: {self.model_server}")
            
            # Return in sagemaker-core compatible format
            return InvokeEndpointOutput(
                body=io.BytesIO(json.dumps(response_data).encode('utf-8')),
                content_type=accept
            )
        

    @classmethod
    def create(
        cls,
        endpoint_name: str,
        endpoint_config_name: Optional[str] = None,
        local_model: Optional[Model] = None,
        local_session=None,
        in_process_mode=False, 
        local_container_mode_obj=None,
        in_process_mode_obj=None,
        model_server=None,
        secret_key=None,
        serializer=None,
        deserializer=None,
        container_config="auto",
        **kwargs
    ) -> "LocalEndpoint":
        """Create and start local endpoint."""
        if local_session is None:
            from sagemaker.core.local.local_session import LocalSession
            local_session = LocalSession()
        

        if in_process_mode:
            endpoint = cls(
                endpoint_name=endpoint_name,
                endpoint_config_name=endpoint_config_name or f"{endpoint_name}-config",
                local_session=local_session,
                local_model=local_model,
                in_process_mode=in_process_mode,
                local_container_mode_obj=local_container_mode_obj,
                in_process_mode_obj=in_process_mode_obj,
                model_server=model_server,
                secret_key=secret_key,
                serializer=serializer,
                deserializer=deserializer,
                container_config=container_config,
                **kwargs
            )

            endpoint.in_process_mode_obj.create_server(
                ping_fn=endpoint._universal_deep_ping
            )

            return endpoint
            

        else:
            # Create endpoint instance first so we can reference its ping method
            endpoint = cls(
                endpoint_name=endpoint_name,
                endpoint_config_name=endpoint_config_name or f"{endpoint_name}-config",
                local_session=local_session,
                local_model=local_model,
                in_process_mode=in_process_mode,
                local_container_mode_obj=local_container_mode_obj,
                in_process_mode_obj=in_process_mode_obj,
                model_server=model_server,
                secret_key=secret_key,
                serializer=serializer,
                deserializer=deserializer,
                container_config=container_config,
                **kwargs
            )
            
            # Start container with ping function
            endpoint.local_container_mode_obj.create_server(
                image=local_model.primary_container.image,
                container_timeout_seconds=kwargs.get("container_timeout_seconds", 300),
                secret_key=endpoint.secret_key,
                ping_fn=endpoint._universal_deep_ping,
                env_vars=local_model.primary_container.environment or {},
                model_path=endpoint.local_container_mode_obj.model_path,
                container_config=_get_container_config(endpoint.container_config)
            )
            
            # Register endpoint with V2 LocalSession
            production_variants = [{
                "VariantName": "AllTraffic",
                "ModelName": local_model.model_name, 
                "InitialInstanceCount": 1,
                "InstanceType": "local"
            }]

            local_session.sagemaker_client.create_endpoint_config(
                EndpointConfigName=endpoint.endpoint_config_name,
                ProductionVariants=production_variants
            )

            # Then create endpoint
            local_session.sagemaker_client.create_endpoint(
                EndpointName=endpoint_name,
                EndpointConfigName=endpoint.endpoint_config_name  
            )
            
            return endpoint


    @classmethod
    def get(cls, endpoint_name: str, local_session=None) -> Optional["LocalEndpoint"]:
        """Get existing local endpoint.
        
        Implementation based on V2 LocalSession.describe_endpoint()
        Reference: /sagemaker/local/local_session.py:describe_endpoint()
        """
        if local_session is None:
            from sagemaker.core.local.local_session import LocalSession
            local_session = LocalSession()
        
        try:
            # Call V2 describe_endpoint to get endpoint info
            endpoint_info = local_session.sagemaker_client.describe_endpoint(
                EndpointName=endpoint_name
            )
            
            return cls(
                endpoint_name=endpoint_name,
                endpoint_config_name=endpoint_info["EndpointConfigName"],
                local_session=local_session
            )
        except Exception:
            # Endpoint not found
            return None
    
    def refresh(self) -> "LocalEndpoint":
        """Refresh endpoint state.
        
        Implementation based on V2 LocalSession.describe_endpoint()
        Reference: /sagemaker/local/local_session.py:describe_endpoint()
        """
        endpoint_info = self._local_session.sagemaker_client.describe_endpoint(
            EndpointName=self.endpoint_name
        )
        
        # Update attributes from V2 response
        self.endpoint_config_name = endpoint_info["EndpointConfigName"]
        
        return self
    
    def delete(self) -> None:
        """Delete local endpoint and cleanup container.
        
        Implementation based on V2 LocalSession.delete_endpoint()
        Reference: /sagemaker/local/local_session.py:delete_endpoint()
        This calls _LocalEndpoint.stop() which stops the Docker container
        """
        self._local_session.sagemaker_client.delete_endpoint(
            EndpointName=self.endpoint_name
        )
    
    def update(self, endpoint_config_name: str) -> None:
        """Update endpoint configuration.
        
        V2 Reference: /sagemaker/local/local_session.py:update_endpoint()
        Note: V2 raises NotImplementedError for update_endpoint
        """
        raise NotImplementedError("Update endpoint is not supported in local mode")


class LocalEndpointConfig:
    """Local endpoint configuration that mimics sagemaker.core.EndpointConfig interface."""
    
    def __init__(
        self,
        endpoint_config_name: str,
        production_variants: list,
        local_session=None,
        **kwargs
    ):
        """Initialize local endpoint config.
        
        Args:
            endpoint_config_name: Name of the endpoint configuration
            production_variants: List of production variant configurations
            local_session: V2 LocalSession instance
        """
        self.endpoint_config_name = endpoint_config_name
        self.production_variants = production_variants
        self.creation_time = datetime.datetime.now()
        
        if local_session is None:
            from sagemaker.core.local.local_session import LocalSession
            self._local_session = LocalSession()
        else:
            self._local_session = local_session
    
    @classmethod
    def create(
        cls,
        endpoint_config_name: str,
        production_variants: list,
        local_session=None,
        **kwargs
    ) -> "LocalEndpointConfig":
        """Create local endpoint configuration.
        
        Implementation based on V2 LocalSession.create_endpoint_config()
        Reference: /sagemaker/local/local_session.py:create_endpoint_config()
        """
        if local_session is None:
            from sagemaker.core.local.local_session import LocalSession
            local_session = LocalSession()
        
        # Create instance
        local_config = cls(
            endpoint_config_name=endpoint_config_name,
            production_variants=production_variants,
            local_session=local_session
        )
        
        # Call V2 LocalSession.create_endpoint_config()
        local_session.sagemaker_client.create_endpoint_config(
            EndpointConfigName=endpoint_config_name,
            ProductionVariants=production_variants
        )
        
        return local_config
    
    def delete(self) -> None:
        """Delete local endpoint configuration.
        
        Implementation based on V2 LocalSession.delete_endpoint_config()
        Reference: /sagemaker/local/local_session.py:delete_endpoint_config()
        """
        self._local_session.sagemaker_client.delete_endpoint_config(
            EndpointConfigName=self.endpoint_config_name
        )



def _get_container_config(config: str) -> dict:
    """Get container configuration based on config type."""
    if config == "host":
        return {"network_mode": "host"}
    elif config == "bridge":
        return {"ports": {'8080/tcp': 8080}}
    elif config == "auto":
        import platform
        if platform.system().lower() == "linux":
            return {"network_mode": "host"}
        else:
            return {"ports": {'8080/tcp': 8080}}
    else:
        raise ValueError("container_config must be 'host', 'bridge', or 'auto'")


# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Placeholder docstring"""
from __future__ import absolute_import
import threading
import time
import uuid
from botocore.exceptions import WaiterError

from sagemaker.core import s3
from sagemaker.core.exceptions import PollingTimeoutError, AsyncInferenceModelError
from sagemaker.serve.async_inference import WaiterConfig, AsyncInferenceResponse
from sagemaker.core.s3 import parse_s3_url
from sagemaker.core.helper.session_helper import Session
from sagemaker.core.common_utils import name_from_base, sagemaker_timestamp, format_tags


class AsyncPredictor:
    """Make async prediction requests to an Amazon SageMaker endpoint."""

    def __init__(
        self,
        predictor,
        name=None,
    ):
        """Initialize an ``AsyncPredictor``.

        Args:
            predictor (sagemaker.predictor.Predictor): General ``Predictor``
                object has useful methods and variables. ``AsyncPredictor``
                stands on top of it with capability for async inference.
        """
        self.predictor = predictor
        self.endpoint_name = predictor.endpoint_name
        self.sagemaker_session = predictor.sagemaker_session or Session()
        if self.sagemaker_session.s3_client is None:
            self.s3_client = self.sagemaker_session.boto_session.client(
                "s3",
                region_name=self.sagemaker_session.boto_region_name,
            )
        else:
            self.s3_client = self.sagemaker_session.s3_client

        self.serializer = predictor.serializer
        self.deserializer = predictor.deserializer
        self.name = name
        self._endpoint_config_name = None
        self._model_names = None
        self._context = None
        self._input_path = None

    def predict(
        self,
        data=None,
        input_path=None,
        initial_args=None,
        inference_id=None,
        waiter_config=WaiterConfig(),
    ):
        """Wait and return the Async Inference result from the specified endpoint.

        Args:
            data (object): Input data for which you want the model to provide
                inference. If a serializer was specified in the encapsulated
                Predictor object, the result of the serializer is sent as input
                data. Otherwise the data must be sequence of bytes, and the
                predict method then sends the bytes in the request body as is.
            input_path (str): Amazon S3 URI contains input data for which you want
                the model to provide async inference. (Default: None)
            initial_args (dict[str,str]): Optional. Default arguments for boto3
                ``invoke_endpoint_async`` call. (Default: None).
            inference_id (str): If you provide a value, it is added to the captured data
                when you enable data capture on the endpoint (Default: None).
            waiter_config (sagemaker.async_inference.waiter_config.WaiterConfig): Configuration
                for the waiter. (Default: {"Delay": 15 seconds, "MaxAttempts": 60}
        Raises:
            ValueError: If both input data and input Amazon S3 path are not provided
        Returns:
            object: Inference for the given input. If a deserializer was specified when creating
                the Predictor, the result of the deserializer is
                returned. Otherwise the response returns the sequence of bytes
                as is.
        """
        if data is None and input_path is None:
            raise ValueError(
                "Please provide input data or input Amazon S3 location to use async prediction"
            )
        if data is not None:
            input_path = self._upload_data_to_s3(data, input_path)

        self._input_path = input_path
        response = self._submit_async_request(input_path, initial_args, inference_id)
        output_location = response["OutputLocation"]
        failure_location = response.get("FailureLocation")
        result = self._wait_for_output(
            output_path=output_location, failure_path=failure_location, waiter_config=waiter_config
        )

        return result

    def predict_async(
        self,
        data=None,
        input_path=None,
        initial_args=None,
        inference_id=None,
    ):
        """Return the Async Inference ouput Amazon S3 path from the specified endpoint.

        Args:
            data (object): Input data for which you want the model to provide
                inference. If a serializer was specified in the encapsulated
                Predictor object, the result of the serializer is sent as input
                data. Otherwise the data must be sequence of bytes, and the
                predict method then upload the data to the ``input_s3_path``. If
                ``input_s3_path`` is None, upload the data to
            input_path (str): Amazon S3 URI contains input data for which you want
                the model to provide async inference. (Default: None)
            initial_args (dict[str,str]): Optional. Default arguments for boto3
                ``invoke_endpoint_async`` call. (Default: None).
            inference_id (str): If you provide a value, it is added to the captured data
                when you enable data capture on the endpoint (Default: None).
        Raises:
            ValueError: If both input data and input Amazon S3 path are not provided
        Returns:
            AsyncInferenceResponse:
                Inference response for the given input. It provides method to check
                the result in the Amazon S3 output path.
        """
        if data is None and input_path is None:
            raise ValueError(
                "Please provide input data or input Amazon S3 location to use async prediction"
            )
        if data is not None:
            input_path = self._upload_data_to_s3(data, input_path)

        self._input_path = input_path
        response = self._submit_async_request(input_path, initial_args, inference_id)
        output_location = response["OutputLocation"]
        failure_location = response.get("FailureLocation")
        response_async = AsyncInferenceResponse(
            predictor_async=self,
            output_path=output_location,
            failure_path=failure_location,
        )

        return response_async

    def _upload_data_to_s3(
        self,
        data,
        input_path=None,
    ):
        """Upload request data to Amazon S3 for users"""
        if input_path:
            bucket, key = parse_s3_url(input_path)
        else:
            my_uuid = str(uuid.uuid4())
            timestamp = sagemaker_timestamp()
            bucket = self.sagemaker_session.default_bucket()
            key = s3.s3_path_join(
                self.sagemaker_session.default_bucket_prefix,
                "async-endpoint-inputs",
                name_from_base(self.name, short=True),
                "{}-{}".format(timestamp, my_uuid),
            )

        data = self.serializer.serialize(data)
        self.s3_client.put_object(
            Body=data, Bucket=bucket, Key=key, ContentType=self.serializer.CONTENT_TYPE
        )
        input_path = input_path or "s3://{}/{}".format(bucket, key)

        return input_path

    def _create_request_args(
        self,
        input_path,
        initial_args=None,
        inference_id=None,
    ):
        """Create request args for ``invoke_endpoint_async``"""
        args = dict(initial_args) if initial_args else {}
        args["InputLocation"] = input_path
        if "EndpointName" not in args:
            args["EndpointName"] = self.predictor.endpoint_name

        if "Accept" not in args:
            args["Accept"] = ", ".join(self.predictor.accept)

        if inference_id:
            args["InferenceId"] = inference_id

        return args

    def _submit_async_request(
        self,
        input_path,
        initial_args,
        inference_id,
    ):
        """Create request and invoke async endpoint with the request"""
        request_args = self._create_request_args(input_path, initial_args, inference_id)

        response = self.sagemaker_session.sagemaker_runtime_client.invoke_endpoint_async(
            **request_args
        )

        return response

    def _wait_for_output(self, output_path, failure_path, waiter_config):
        """Retrieve output based on the presense of failure_path."""
        if failure_path is not None:
            return self._check_output_and_failure_paths(output_path, failure_path, waiter_config)

        return self._check_output_path(output_path, waiter_config)

    def _check_output_path(self, output_path, waiter_config):
        """Check the Amazon S3 output path for the output.

        Periodically check Amazon S3 output path for async inference result.
        Timeout automatically after max attempts reached
        """
        bucket, key = parse_s3_url(output_path)
        s3_waiter = self.s3_client.get_waiter("object_exists")
        try:
            s3_waiter.wait(Bucket=bucket, Key=key, WaiterConfig=waiter_config._to_request_dict())
        except WaiterError:
            raise PollingTimeoutError(
                message="Inference could still be running",
                output_path=output_path,
                seconds=waiter_config.delay * waiter_config.max_attempts,
            )
        s3_object = self.s3_client.get_object(Bucket=bucket, Key=key)
        result = self.predictor._handle_response(response=s3_object)
        return result

    def _check_output_and_failure_paths(self, output_path, failure_path, waiter_config):
        """Check the Amazon S3 output path for the output.

        This method waits for either the output file or the failure file to be found on the
        specified S3 output path. Whichever file is found first, its corresponding event is
        triggered, and the method executes the appropriate action based on the event.

        Args:
            output_path (str): The S3 path where the output file is expected to be found.
            failure_path (str): The S3 path where the failure file is expected to be found.
            waiter_config (boto3.waiter.WaiterConfig): The configuration for the S3 waiter.

        Returns:
            Any: The deserialized result from the output file, if the output file is found first.
            Otherwise, raises an exception.

        Raises:
            AsyncInferenceModelError: If the failure file is found before the output file.
            PollingTimeoutError: If both files are not found and the S3 waiter
             has thrown a WaiterError.
        """
        output_bucket, output_key = parse_s3_url(output_path)
        failure_bucket, failure_key = parse_s3_url(failure_path)

        output_file_found = threading.Event()
        failure_file_found = threading.Event()

        def check_output_file():
            try:
                output_file_waiter = self.s3_client.get_waiter("object_exists")
                output_file_waiter.wait(
                    Bucket=output_bucket,
                    Key=output_key,
                    WaiterConfig=waiter_config._to_request_dict(),
                )
                output_file_found.set()
            except WaiterError:
                pass

        def check_failure_file():
            try:
                failure_file_waiter = self.s3_client.get_waiter("object_exists")
                failure_file_waiter.wait(
                    Bucket=failure_bucket,
                    Key=failure_key,
                    WaiterConfig=waiter_config._to_request_dict(),
                )
                failure_file_found.set()
            except WaiterError:
                pass

        output_thread = threading.Thread(target=check_output_file)
        failure_thread = threading.Thread(target=check_failure_file)

        output_thread.start()
        failure_thread.start()

        while not output_file_found.is_set() and not failure_file_found.is_set():
            time.sleep(1)

        if output_file_found.is_set():
            s3_object = self.s3_client.get_object(Bucket=output_bucket, Key=output_key)
            result = self.predictor._handle_response(response=s3_object)
            return result

        failure_object = self.s3_client.get_object(Bucket=failure_bucket, Key=failure_key)
        failure_response = self.predictor._handle_response(response=failure_object)

        raise (
            AsyncInferenceModelError(message=failure_response)
            if failure_file_found.is_set()
            else PollingTimeoutError(
                message="Inference could still be running",
                output_path=output_path,
                seconds=waiter_config.delay * waiter_config.max_attempts,
            )
        )

    def update_endpoint(
        self,
        initial_instance_count=None,
        instance_type=None,
        accelerator_type=None,
        model_name=None,
        tags=None,
        kms_key=None,
        data_capture_config_dict=None,
        wait=True,
    ):
        """Update the existing endpoint with the provided attributes.

        This creates a new EndpointConfig in the process. If ``initial_instance_count``,
        ``instance_type``, ``accelerator_type``, or ``model_name`` is specified, then a new
        ProductionVariant configuration is created; values from the existing configuration
        are not preserved if any of those parameters are specified.

        Args:
            initial_instance_count (int): The initial number of instances to run in the endpoint.
                This is required if ``instance_type``, ``accelerator_type``, or ``model_name`` is
                specified. Otherwise, the values from the existing endpoint configuration's
                ProductionVariants are used.
            instance_type (str): The EC2 instance type to deploy the endpoint to.
                This is required if ``initial_instance_count`` or ``accelerator_type`` is specified.
                Otherwise, the values from the existing endpoint configuration's
                ``ProductionVariants`` are used.
            accelerator_type (str): The type of Elastic Inference accelerator to attach to
                the endpoint, e.g. "ml.eia1.medium". If not specified, and
                ``initial_instance_count``, ``instance_type``, and ``model_name`` are also ``None``,
                the values from the existing endpoint configuration's ``ProductionVariants`` are
                used. Otherwise, no Elastic Inference accelerator is attached to the endpoint.
            model_name (str): The name of the model to be associated with the endpoint.
                This is required if ``initial_instance_count``, ``instance_type``, or
                ``accelerator_type`` is specified and if there is more than one model associated
                with the endpoint. Otherwise, the existing model for the endpoint is used.
            tags (list[dict[str, str]]): The list of tags to add to the endpoint
                config. If not specified, the tags of the existing endpoint configuration are used.
                If any of the existing tags are reserved AWS ones (i.e. begin with "aws"),
                they are not carried over to the new endpoint configuration.
            kms_key (str): The KMS key that is used to encrypt the data on the storage volume
                attached to the instance hosting the endpoint If not specified,
                the KMS key of the existing endpoint configuration is used.
            data_capture_config_dict (dict): The endpoint data capture configuration
                for use with Amazon SageMaker Model Monitoring. If not specified,
                the data capture configuration of the existing endpoint configuration is used.
            wait (bool): Wait for updating to finish
        """

        self.predictor.update_endpoint(
            initial_instance_count=initial_instance_count,
            instance_type=instance_type,
            accelerator_type=accelerator_type,
            model_name=model_name,
            tags=format_tags(tags),
            kms_key=kms_key,
            data_capture_config_dict=data_capture_config_dict,
            wait=wait,
        )

    def delete_endpoint(self, delete_endpoint_config=True):
        """Delete the Amazon SageMaker endpoint backing this async predictor.

        This also delete the endpoint configuration attached to it if
        delete_endpoint_config is True.

        Args:
            delete_endpoint_config (bool, optional): Flag to indicate whether to
                delete endpoint configuration together with endpoint. Defaults
                to True. If True, both endpoint and endpoint configuration will
                be deleted. If False, only endpoint will be deleted.
        """
        self.predictor.delete_endpoint(delete_endpoint_config)

    def delete_model(self):
        """Deletes the Amazon SageMaker models backing this predictor."""
        self.predictor.delete_model()

    def enable_data_capture(self):
        """Enables data capture by updating DataCaptureConfig.

        This function updates the DataCaptureConfig for the Predictor's associated Amazon SageMaker
        Endpoint to enable data capture. For a more customized experience, refer to
        update_data_capture_config, instead.
        """
        self.predictor.enable_data_capture()

    def disable_data_capture(self):
        """Disables data capture by updating DataCaptureConfig.

        This function updates the DataCaptureConfig for the Predictor's associated Amazon SageMaker
        Endpoint to disable data capture. For a more customized experience, refer to
        update_data_capture_config, instead.
        """
        self.predictor.disable_data_capture()

    def update_data_capture_config(self, data_capture_config):
        """Updates the DataCaptureConfig for the Predictor's associated Amazon SageMaker Endpoint.

        Update is done using the provided DataCaptureConfig.

        Args:
            data_capture_config (sagemaker.model_monitor.DataCaptureConfig): The
                DataCaptureConfig to update the predictor's endpoint to use.
        """
        self.predictor.update_data_capture_config(data_capture_config)

    def list_monitors(self):
        """Generates ModelMonitor objects (or DefaultModelMonitors).

        Objects are generated based on the schedule(s) associated with the endpoint
        that this predictor refers to.

        Returns:
            [sagemaker.model_monitor.model_monitoring.ModelMonitor]: A list of
                ModelMonitor (or DefaultModelMonitor) objects.

        """
        return self.predictor.list_monitors()

    def endpoint_context(self):
        """Retrieves the lineage context object representing the endpoint.

        Examples:
            .. code-block:: python

                predictor = Predictor()
                context = predictor.endpoint_context()
                models = context.models()

        Returns:
            ContextEndpoint: The context for the endpoint.
        """

        return self.predictor.endpoint_context()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds templated classes to enable users to provide custom inference scripting capabilities"""
from __future__ import absolute_import
from abc import ABC, abstractmethod


class CustomOrchestrator(ABC):
    """Templated class to standardize sync entrypoint-based inference scripts"""

    def __init__(self):
        self._client = None

    @property
    def client(self):
        """Boto3 SageMaker runtime client to use with custom orchestrator"""
        if not hasattr(self, "_client") or not self._client:
            from boto3 import Session

            self._client = Session().client("sagemaker-runtime")
        return self._client

    @abstractmethod
    def handle(self, data, context=None):
        """Abstract class for defining an entrypoint for the model server"""
        return NotImplemented


class AsyncCustomOrchestrator(ABC):
    """Templated class to standardize async entrypoint-based inference scripts"""

    @abstractmethod
    async def handle(self, data, context=None):
        """Abstract class for defining an aynchronous entrypoint for the model server"""
        return NotImplemented



"""DJL ModelBuilder Utils"""

from __future__ import absolute_import
import math
import logging
from sagemaker.serve.utils.local_hardware import _get_available_gpus
from sagemaker.serve.builder.schema_builder import SchemaBuilder

logger = logging.getLogger(__name__)

ATTENTION_HEAD_NAME_VARIENTS = ["n_head", "n_heads", "num_head", "num_heads", "num_attention_heads"]
CHARS_PER_TOKEN = 4
TOKENS_PER_WORD = 0.75


def _get_default_tensor_parallel_degree(hf_model_config: dict, gpu_count: int = None) -> int:
    """Placeholder docstring"""
    available_gpus = _get_available_gpus()
    if not available_gpus and not gpu_count:
        return None

    attention_heads = None
    for variant in ATTENTION_HEAD_NAME_VARIENTS:
        attention_heads = hf_model_config.get(variant)
        if attention_heads:
            break

    if not attention_heads:
        return 1

    tot_gpus = len(available_gpus) if available_gpus else gpu_count
    for i in (n + 1 for n in reversed(range(tot_gpus))):
        if attention_heads % i == 0:
            logger.info(
                "Max GPU parallelism of %s is allowed. Total attention heads %s", i, attention_heads
            )
            return i

    return 1


def _get_default_data_type() -> tuple:
    """Placeholder docstring"""
    return "bf16"


def _get_default_batch_size() -> int:
    """Placeholder docstring"""
    return 1


def _set_tokens_to_tokens_threshold(tokens: int) -> int:
    """Placeholder docstring"""
    if tokens <= 128:
        return 128
    if tokens <= 256:
        return 256
    elif tokens <= 512:
        return 512
    elif tokens <= 1024:
        return 1024
    elif tokens <= 2048:
        return 2048
    return 4096


def _tokens_from_chars(text: str) -> int:
    """Placeholder docstring"""
    return len(text) / CHARS_PER_TOKEN


def _tokens_from_words(text: str) -> int:
    """Placeholder docstring"""
    return math.ceil(len(text.split(" ")) * TOKENS_PER_WORD)


def _get_default_max_tokens(sample_input, sample_output) -> tuple:
    """Placeholder docstring"""
    inputs = sample_input.get("inputs")
    generated_text = sample_output[0].get("generated_text")

    input_tokens_from_chars = _tokens_from_chars(inputs)
    input_tokens_from_words = _tokens_from_words(inputs)
    max_input_tokens = max(input_tokens_from_chars, input_tokens_from_words)

    output_tokens_from_chars = _tokens_from_chars(generated_text)
    output_tokens_from_words = _tokens_from_words(generated_text)
    max_output_tokens = max(output_tokens_from_chars, output_tokens_from_words)

    max_total_tokens = _set_tokens_to_tokens_threshold(max_input_tokens + max_output_tokens)

    max_new_tokens = sample_input.get("parameters").get("max_new_tokens")
    if not max_new_tokens:
        max_new_tokens = _set_tokens_to_tokens_threshold(max_output_tokens)

    return (max_total_tokens, max_new_tokens)


def _get_default_djl_configurations(
    model_id: str, hf_model_config: dict, schema_builder: SchemaBuilder
) -> tuple:
    """Placeholder docstring"""
    default_tensor_parallel_degree = _get_default_tensor_parallel_degree(hf_model_config)
    if default_tensor_parallel_degree is None:
        default_tensor_parallel_degree = "max"
    default_data_type = _get_default_data_type()
    default_max_tokens, default_max_new_tokens = _get_default_max_tokens(
        schema_builder.sample_input, schema_builder.sample_output
    )

    env = {
        "TENSOR_PARALLEL_DEGREE": str(default_tensor_parallel_degree),
        "OPTION_DTYPE": default_data_type,
    }
    return (env, default_max_new_tokens)


def _get_admissible_tensor_parallel_degrees(hf_model_config: dict) -> int:
    """Placeholder docstring"""
    available_gpus = _get_available_gpus()

    attention_heads = None
    for variant in ATTENTION_HEAD_NAME_VARIENTS:
        attention_heads = hf_model_config.get(variant)
        if attention_heads:
            break

    if not attention_heads:
        return 1

    admissible_parallel_degrees = []
    for i in (n + 1 for n in reversed(range(len(available_gpus)))):
        if attention_heads % i == 0:
            admissible_parallel_degrees.append(i)

    logger.info("Model can be sharded across %s GPUs", admissible_parallel_degrees)

    return admissible_parallel_degrees


def _get_admissible_dtypes():
    """Placeholder docstring"""
    return ["bf16"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Placeholder docstring"""
from __future__ import absolute_import
import logging
from time import perf_counter

import requests

from sagemaker.core.helper.session_helper import Session, exceptions
from sagemaker.serve.mode.function_pointers import Mode
from sagemaker.serve.model_format.mlflow.constants import MLFLOW_MODEL_PATH
from sagemaker.serve.utils.exceptions import ModelBuilderException
from sagemaker.serve.utils.lineage_constants import (
    MLFLOW_LOCAL_PATH,
    MLFLOW_S3_PATH,
    MLFLOW_MODEL_PACKAGE_PATH,
    MLFLOW_RUN_ID,
    MLFLOW_REGISTRY_PATH,
)
from sagemaker.serve.utils.lineage_utils import _get_mlflow_model_path_type
from sagemaker.serve.utils.types import (
    ModelServer,
    ImageUriOption,
    ModelHub,
    SpeculativeDecodingDraftModelSource,
)
from sagemaker.serve.validations.check_image_uri import is_1p_image_uri
from sagemaker.core.user_agent import SDK_VERSION

logger = logging.getLogger(__name__)

TELEMETRY_OPT_OUT_MESSAGING = (
    "ModelBuilder will collect telemetry to help us better understand our user's needs, "
    "diagnose issues, and deliver additional features. "
    "To opt out of telemetry, please disable "
    "via TelemetryOptOut in intelligent defaults. See "
    "https://sagemaker.readthedocs.io/en/stable/overview.html#"
    "configuring-and-using-defaults-with-the-sagemaker-python-sdk "
    "for more info."
)

MODE_TO_CODE = {
    str(Mode.IN_PROCESS): 1,
    str(Mode.LOCAL_CONTAINER): 2,
    str(Mode.SAGEMAKER_ENDPOINT): 3,
}

MODEL_SERVER_TO_CODE = {
    str(ModelServer.TORCHSERVE): 1,
    str(ModelServer.MMS): 2,
    str(ModelServer.TENSORFLOW_SERVING): 3,
    str(ModelServer.DJL_SERVING): 4,
    str(ModelServer.TRITON): 5,
    str(ModelServer.TGI): 6,
    str(ModelServer.TEI): 7,
    str(ModelServer.SMD): 8,
}

MLFLOW_MODEL_PATH_CODE = {
    MLFLOW_LOCAL_PATH: 1,
    MLFLOW_S3_PATH: 2,
    MLFLOW_MODEL_PACKAGE_PATH: 3,
    MLFLOW_RUN_ID: 4,
    MLFLOW_REGISTRY_PATH: 5,
}

MODEL_HUB_TO_CODE = {
    str(ModelHub.JUMPSTART): 1,
    str(ModelHub.HUGGINGFACE): 2,
}

SD_DRAFT_MODEL_SOURCE_TO_CODE = {
    str(SpeculativeDecodingDraftModelSource.SAGEMAKER): 1,
    str(SpeculativeDecodingDraftModelSource.CUSTOM): 2,
}


def _capture_telemetry(func_name: str):
    """Placeholder docstring"""

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            # Call the original function
            logger.info(TELEMETRY_OPT_OUT_MESSAGING)
            response = None
            caught_ex = None
            status = "1"
            failure_reason = None
            failure_type = None
            extra = f"{func_name}"

            start_timer = perf_counter()
            try:
                response = func(self, *args, **kwargs)
            except (
                ModelBuilderException,
                exceptions.CapacityError,
                exceptions.UnexpectedStatusException,
                exceptions.AsyncInferenceError,
            ) as e:
                status = "0"
                caught_ex = e
                failure_reason = str(e)
                failure_type = e.__class__.__name__
            except Exception as e:  # pylint: disable=W0703
                raise e

            stop_timer = perf_counter()
            elapsed = stop_timer - start_timer

            if self.model_server:
                extra += f"&x-modelServer={MODEL_SERVER_TO_CODE[str(self.model_server)]}"

            if self.image_uri:
                image_uri_option = _get_image_uri_option(
                    self.image_uri, getattr(self, "_is_custom_image_uri", False)
                )
                split_image_uri = self.image_uri.split("/")
                if len(split_image_uri) > 1:
                    extra += f"&x-imageTag={split_image_uri[1]}"

            extra += f"&x-sdkVersion={SDK_VERSION}"

            if self.image_uri:
                extra += f"&x-defaultImageUsage={image_uri_option}"

            if self.model_server == ModelServer.DJL_SERVING or self.model_server == ModelServer.TGI:
                extra += f"&x-modelName={self.model}"

            if self.sagemaker_session and self.sagemaker_session.endpoint_arn:
                extra += f"&x-endpointArn={self.sagemaker_session.endpoint_arn}"

            if getattr(self, "_is_mlflow_model", False):
                mlflow_model_path = self.model_metadata[MLFLOW_MODEL_PATH]
                mlflow_model_path_type = _get_mlflow_model_path_type(mlflow_model_path)
                extra += f"&x-mlflowModelPathType={MLFLOW_MODEL_PATH_CODE[mlflow_model_path_type]}"

            if getattr(self, "model_hub", False):
                extra += f"&x-modelHub={MODEL_HUB_TO_CODE[str(self.model_hub)]}"

            if getattr(self, "is_fine_tuned", False):
                extra += "&x-fineTuned=1"

            if getattr(self, "is_compiled", False):
                extra += "&x-compiled=1"
            if getattr(self, "is_quantized", False):
                extra += "&x-quantized=1"
            if getattr(self, "speculative_decoding_draft_model_source", False):
                model_provider_enum = (
                    SpeculativeDecodingDraftModelSource.SAGEMAKER
                    if self.speculative_decoding_draft_model_source == "sagemaker"
                    else SpeculativeDecodingDraftModelSource.CUSTOM
                )
                model_provider_value = SD_DRAFT_MODEL_SOURCE_TO_CODE[str(model_provider_enum)]
                extra += f"&x-sdDraftModelSource={model_provider_value}"

            if getattr(self, "deployment_config_name", False):
                config_name_code = self.deployment_config_name.lower()
                extra += f"&x-configName={config_name_code}"

            extra += f"&x-latency={round(elapsed, 2)}"

            if hasattr(self, "serve_settings") and not self.serve_settings.telemetry_opt_out:
                _send_telemetry(
                    status,
                    MODE_TO_CODE[str(self.mode)],
                    self.sagemaker_session,
                    failure_reason,
                    failure_type,
                    extra,
                )

            if caught_ex:
                raise caught_ex

            return response

        return wrapper

    return decorator


def _send_telemetry(
    status: str,
    mode: int,
    session: Session,
    failure_reason: str = None,
    failure_type: str = None,
    extra_info: str = None,
) -> None:
    """Make GET request to an empty object in S3 bucket"""
    try:
        accountId = _get_accountId(session)
        region = _get_region_or_default(session)
        url = _construct_url(
            accountId,
            str(mode),
            status,
            failure_reason,
            failure_type,
            extra_info,
            region,
        )
        _requests_helper(url, 2)
        logger.debug("ModelBuilder metrics emitted.")
    except Exception:  # pylint: disable=W0703
        logger.debug("ModelBuilder metrics not emitted")


def _construct_url(
    accountId: str,
    mode: str,
    status: str,
    failure_reason: str,
    failure_type: str,
    extra_info: str,
    region: str,
) -> str:
    """Placeholder docstring"""

    base_url = (
        f"https://sm-pysdk-t-{region}.s3.{region}.amazonaws.com/telemetry?"
        f"x-accountId={accountId}"
        f"&x-mode={mode}"
        f"&x-status={status}"
    )
    if failure_reason:
        base_url += f"&x-failureReason={failure_reason}"
        base_url += f"&x-failureType={failure_type}"
    if extra_info:
        base_url += f"&x-extra={extra_info}"
    return base_url


def _requests_helper(url, timeout):
    """Placeholder docstring"""

    response = None
    try:
        response = requests.get(url, timeout)
    except requests.exceptions.RequestException as e:
        logger.debug("Request exception: %s", str(e))
    return response


def _get_accountId(session):
    """Placeholder docstring"""

    try:
        sts = session.boto_session.client("sts")
        return sts.get_caller_identity()["Account"]
    except Exception:  # pylint: disable=W0703
        return None


def _get_region_or_default(session):
    """Placeholder docstring"""

    try:
        return session.boto_session.region_name
    except Exception:  # pylint: disable=W0703
        return "us-west-2"


def _get_image_uri_option(image_uri: str, is_custom_image: bool) -> int:
    """Detect whether default values are used for ModelBuilder

    Args:
        image_uri (str): Image uri used by ModelBuilder.
        is_custom_image: (bool): Boolean indicating whether customer provides with custom image.
    Returns:
        bool: Integer code of image option types.
    """

    if not is_custom_image:
        return ImageUriOption.DEFAULT_IMAGE.value

    if is_1p_image_uri(image_uri):
        return ImageUriOption.CUSTOM_1P_IMAGE.value

    return ImageUriOption.CUSTOM_IMAGE.value



"""Module for Local Tensorflow Server"""

from __future__ import absolute_import

import requests
import logging
import platform
from pathlib import Path
from sagemaker.core.common_utils import _is_s3_uri
from sagemaker.core.helper.session_helper import Session
from sagemaker.core.s3.utils import determine_bucket_and_prefix, parse_s3_url
from sagemaker.core import fw_utils
from sagemaker.serve.utils.uploader import upload
from sagemaker.core.local.local_session import get_docker_host

logger = logging.getLogger(__name__)


class LocalTensorflowServing:
    """LocalTensorflowServing class."""

    def _start_tensorflow_serving(
        self, client: object, image: str, model_path: str, secret_key: str, env_vars: dict
    ):
        """Starts a local tensorflow serving container.

        Args:
            client: Docker client
            image: Image to use
            model_path: Path to the model
            secret_key: Secret key to use for authentication
            env_vars: Environment variables to set
        """
        self.container = client.containers.run(
            image,
            "serve",
            detach=True,
            auto_remove=False,  # Temporarily disabled to see crash logs
            # network_mode="host",
            ports={'8501/tcp': 8501},
            volumes={
                Path(model_path): {
                    "bind": "/opt/ml/model",
                    "mode": "rw",
                },
            },
            environment={
                "SAGEMAKER_SUBMIT_DIRECTORY": "/opt/ml/model/code",
                "SAGEMAKER_PROGRAM": "inference.py",
                "SAGEMAKER_SERVE_SECRET_KEY": secret_key,
                "LOCAL_PYTHON": platform.python_version(),
                **env_vars,
            },
        )

    def _invoke_tensorflow_serving(self, request: object, content_type: str, accept: str):
        """Invokes a local tensorflow serving container.

        Args:
            request: Request to send
            content_type: Content type to use
            accept: Accept to use
        """
        try:
            response = requests.post(
                f"http://{get_docker_host()}:8501/invocations",
                data=request,
                headers={"Content-Type": content_type, "Accept": accept},
                timeout=60,  # this is what SageMaker Hosting uses as timeout
            )
            response.raise_for_status()
            return response.content
        except Exception as e:
            raise Exception("Unable to send request to the local container server") from e


class SageMakerTensorflowServing:
    """SageMakerTensorflowServing class."""

    def _upload_tensorflow_serving_artifacts(
        self,
        model_path: str,
        sagemaker_session: Session,
        secret_key: str,
        s3_model_data_url: str = None,
        image: str = None,
        should_upload_artifacts: bool = False,
    ):
        """Uploads the model artifacts to S3.

        Args:
            model_path: Path to the model
            sagemaker_session: SageMaker session
            secret_key: Secret key to use for authentication
            s3_model_data_url: S3 model data URL
            image: Image to use
            model_data_s3_path: S3 model data URI
        """
        s3_upload_path = None
        if _is_s3_uri(model_path):
            s3_upload_path = model_path
        elif should_upload_artifacts:
            if s3_model_data_url:
                bucket, key_prefix = parse_s3_url(url=s3_model_data_url)
            else:
                bucket, key_prefix = None, None

            code_key_prefix = fw_utils.model_code_key_prefix(key_prefix, None, image)

            bucket, code_key_prefix = determine_bucket_and_prefix(
                bucket=bucket, key_prefix=code_key_prefix, sagemaker_session=sagemaker_session
            )

            logger.debug(
                "Uploading the model resources to bucket=%s, key_prefix=%s.",
                bucket,
                code_key_prefix,
            )
            s3_upload_path = upload(sagemaker_session, model_path, bucket, code_key_prefix)
            logger.debug("Model resources uploaded to: %s", s3_upload_path)

        env_vars = {
            "SAGEMAKER_SUBMIT_DIRECTORY": "/opt/ml/model/code",
            "SAGEMAKER_PROGRAM": "inference.py",
            "SAGEMAKER_REGION": sagemaker_session.boto_region_name,
            "SAGEMAKER_CONTAINER_LOG_LEVEL": "10",
            "SAGEMAKER_SERVE_SECRET_KEY": secret_key,
            "LOCAL_PYTHON": platform.python_version(),
        }
        return s3_upload_path, env_vars



"""This module is for Triton Python backend."""

from __future__ import absolute_import
import os
import logging
from pathlib import Path
import platform

import triton_python_backend_utils as pb_utils
import cloudpickle
from sagemaker.serve.validations.check_integrity import perform_integrity_check

logger = logging.getLogger(__name__)

TRITON_MODEL_DIR = os.getenv("TRITON_MODEL_DIR")


class TritonPythonModel:
    """A class for Triton Python Backend"""

    @staticmethod
    def auto_complete_config(auto_complete_model_config):
        """Placeholder docstring"""
        return auto_complete_model_config

    def initialize(self, args: dict) -> None:
        """Placeholder docstring"""
        serve_path = Path(TRITON_MODEL_DIR).joinpath("serve.pkl")
        with open(str(serve_path), mode="rb") as f:
            inference_spec, schema_builder = cloudpickle.load(f)

        # TODO: HMAC signing for integrity check

        self.inference_spec = inference_spec
        self.schema_builder = schema_builder
        self.model = inference_spec.load(model_dir=TRITON_MODEL_DIR)

    def execute(self, requests):
        """Placeholder docstring"""
        responses = []

        for request in requests:
            input_ndarray = pb_utils.get_input_tensor_by_name(request, "input_1").as_numpy()
            converted_input = self.schema_builder.input_deserializer.deserialize(input_ndarray)
            output = self.inference_spec.invoke(input_object=converted_input, model=self.model)
            output_ndarray = self.schema_builder.output_serializer.serialize(output)
            response = pb_utils.InferenceResponse(
                output_tensors=[pb_utils.Tensor("output_1", output_ndarray)]
            )
            responses.append(response)

        return responses


def _run_preflight_diagnostics():
    _py_vs_parity_check()
    _pickle_file_integrity_check()


def _py_vs_parity_check():
    container_py_vs = platform.python_version()
    local_py_vs = os.getenv("LOCAL_PYTHON")

    if not local_py_vs or container_py_vs.split(".")[1] != local_py_vs.split(".")[1]:
        logger.warning(
            f"The local python version {local_py_vs} differs from the python version "
            f"{container_py_vs} on the container. Please align the two to avoid unexpected behavior"
        )


def _pickle_file_integrity_check():
    serve_path = Path(TRITON_MODEL_DIR).joinpath("serve.pkl")
    metadata_path = Path(TRITON_MODEL_DIR).joinpath("metadata.json")
    with open(str(serve_path), "rb") as f:
        buffer = f.read()
    perform_integrity_check(buffer=buffer, metadata_path=metadata_path)


# on import, execute
_run_preflight_diagnostics()
